"""
Database stuff!
Should be able to use multiple backends as long as they implement the Python
Database Specification v2.0 (PEP 249).
"""
import sqlite3


def main():
    conn = sqlite3.connect('qed.db')
    c = conn.cursor()
    print(c)

# class Database(object):
#     """
#     Rewritten to just use a database backend.
#     """
#     def __init__(self, db_name, username):
#         self.db_name = db_name
#         self.username = username
#         self.db_conn = None
#         self.cur = None
#
#     def _make_connection(self):
#         self.db_conn = psycopg2.connect(database=self.db_name,
#                                         user=self.username)
#         self.cur = self.db_conn.cursor()
#
#     def _close_connection(self):
#         self.cur.close()
#         self.db_conn.close()
#
#     def __enter__(self):
#         self._make_connection()
#         return self
#
#     def __exit__(self, exc_type, exc_val, exc_tb):
#         self._close_connection()



from pyon.lib.io.parsers import registered_parsers
from pyon.runner.models import Model
from pyon.runner.simulations import Simulation
from pyon.runner.simulations import registered_simulations
from pyon.runner.views import registered_views
from pyon.runner.models import registered_models
from pyon.runner.sources import registered_sources
from pyon.runner import register



from collections import defaultdict, namedtuple
import inspect
from scipy.optimize import minimize
from pyon.lib.register import Register
from pyon.lib.resampling import registered_resamplers
import numpy as np
from pyon.lib.statistics import get_inverse_cov_matrix


__author__ = 'srd1g10'

registered_fitters = {}

FitParams = namedtuple('FitParams', ['average_params', 'errs',
                                     'resampled_params'])
"""
:class:`FitParams` is a namedtuple that packages the things that \
:func:`Fitter.fit` returns.
"""


class Fitter:
    """
    :class:`Fitter` fits data.

    :param fit_func: The function we fit to. Used to calculate the chi squared
    :param resampler: The way we calculate the errors e.g. \
    :class:`Jackknife <resampling.Jackknife>`
    """
    def __init__(self, fit_func=None, resampler='jackknife',
                 resampler_args=None):
        self.fit_func = fit_func
        if resampler_args is None:
            resampler_args = {}
        self.resampler = registered_resamplers[resampler](**resampler_args)

    def fit(self, data=None, errors=None, initial_value=None, fit_range=None,
            covariant=False, correlated=False,
            **kwargs):
        """
        Performs the fit. A specific implementation will use a certain method \
        of minimizing the chi-squared of
        ``fit_func`` against the ``data``.

        :param data: Iterable set of independent measurements that the fit \
        function models.
        :param errors: Iterable set of errors of each measurement.
        :param initial_value: Initial guesses of the best fit of the \
        parameters in ``fit_func``.
        :type initial_value: iterable
        :param fit_range: Iterable set of numbers that determine the data \
        points used in the fit. Can be a list of
            iterables if the fit function requires more than one fit range \
            e.g. simultaneous fit.
        :type fit_range: iterable or list of iterables
        :param covariant: Set the fit as covariant.
        :type covariant: bool
        :param correlated: Set the fit as correlated.
        :type correlated: bool
        """
        resampled_params = defaultdict(list)
        fit_range = np.array(fit_range)
        if covariant:
            pared_data = np.array([[d[t] for d in data] for t in fit_range])
            inv_covar = get_inverse_cov_matrix(pared_data, correlated)
        else:
            inv_covar = None
        for sample in self.resampler.generate_samples(data):
            fit_param = self.fit_one(sample, errors, initial_value, fit_range,
                                     inverse_covariance=inv_covar,
                                     **kwargs)
            for k, v in fit_param.items():
                resampled_params[k].append(v)
        average_params = self.fit_one(np.average(data, axis=0), errors,
                                      initial_value, fit_range,
                                      inverse_covariance=inv_covar, **kwargs)
        errs = self.resampler.calculate_fit_errors(average_params,
                                                   resampled_params)
        return FitParams(average_params, errs, resampled_params)

    def fit_chi_sq(self, **kwargs):
        """
        Implement this in a derived class e.g. :func:`ScipyFitter.fit_chi_sq`.
        """
        raise NotImplementedError("Implement this in a derived class.")

    def fit_one(self, data=None, errors=None, initial_value=None,
                fit_range=None, inverse_covariance=None,
                **kwargs):
        chi_sq = self.generate_chi_sq(data=data, errors=errors,
                                      fit_range=fit_range,
                                      inverse_covariance=inverse_covariance)
        # if hessian:
        #     hess = self._generate_chi_sq_hessian(errors=errors,
        #                                          fit_range=fit_range,
        #                                          hessian=hessian)
        # else:
        #     hess = None
        fit_param = self.fit_chi_sq(chi_sq, initial_value=initial_value,
                                    **kwargs)
        return fit_param

    def generate_chi_sq(self, data, errors=None, inverse_covariance=None,
                        fit_range=None):
        if errors is None:
            errors = [1. for _ in data]
        if inverse_covariance is not None:
            chi_sq = self._generate_chi_sq_covariant(data, inverse_covariance,
                                                     fit_range, self.fit_func)
        else:
            chi_sq = self._generate_chi_sq_uncovariant(data, errors, fit_range,
                                                       self.fit_func)
        return chi_sq

    @staticmethod
    def _generate_chi_sq_covariant(data, inverse_covariance, fit_range,
                                   fit_func):
        def chi_sq(args):
            pared_data = np.array([data[t] for t in fit_range])
            v = np.array(fit_func(fit_range, *args) - pared_data)
            m = np.array(inverse_covariance)
            r = np.dot(m, v)
            c2 = np.dot(v, r)
            return c2 / len(fit_range)
        return chi_sq

    @staticmethod
    def _generate_chi_sq_uncovariant(data, errors, fit_range, fit_func):
        #chi_sq = GenericChi2(fit_func, data, errors, fit_range)
        def chi_sq(args):
            return sum([(data[t] - fit_func(t, *args))**2 / (errors[t])**2
                        for t in fit_range]) / len(fit_range)
        return chi_sq

        # def _generate_chi_sq_hessian(self, errors, fit_range, hessian):
        #     def hess(args):
        #         h_11 = sum([hessian(t, *args)[0]**2 / errors[t]**2
        #                     for t in fit_range])
        #         h_12 = sum([hessian(t, *args)[0]*hessian(t, *args)[1] / \
        # errors[t]**2
        #                     for t in fit_range])
        #         h_21 = h_12
        #         h_22 = sum([hessian(t, *args)[1]**2 / errors[t]**2
        #                     for t in fit_range])
        #         return np.asmatrix([[h_11, h_12], [h_21, h_22]])
        #     return hess




@Register(registered_fitters, 'scipy')
class ScipyFitter(Fitter):
    """
    Inherits from :class:`Fitter`.
    """

    def fit_chi_sq(self, chi_sq, initial_value, method='L-BFGS-B',
                   **kwargs):
        """
        Uses scipy's :func:`minimize <scipy.optimize.minimize>` function to \
        minimize the ``chi_sq`` of the fit.

        :param initial_value: Initial guesses of the best fit of the \
        parameters in ``fit_func``
        :type initial_value: dict
        :param method: Method of fitting to pass through to \
        :func:`minimize <scipy.optimize.minimize>`
        :param \*\*kwargs: Extra parameters to pass in to \
        :func:`minimize <scipy.optimize.minimize>`
        :rtype: dict
        """
        out = minimize(chi_sq, initial_value, method=method,
                       **kwargs)
        fit_params = populate_dict_args(self.fit_func, out.x)
        return fit_params


def populate_dict_args(func, vals):
    """
    A minimization function may return unnamed estimates of the variables. \
    This function maps those values to the
    names of the function arguments. Assume that the first argument of \
    ``func`` the independent variable - we skip it.

    :param func: A function with at least two arguments. The first argument \
    is assumed to be an independent variable
        and is skipped.
    :param vals: Values that are associated with each argument.
    :rtype: dict
    """
    func_args = inspect.getargspec(func).args[1:]
    return {k: v for k, v in zip(func_args, vals)}


def fit_hadron(hadron, initial_value=None, fit_range=None, covariant=False,
               method=None,
               **kwargs):
    """
    Common use case, uses Scipy for fitting and Jackknife for errors.
    """
    if method is None:
        fitter = ScipyFitter(fit_func=hadron.fit_func,
                             resampler='jackknife')
    else:
        fitter = method(fit_func=hadron.fit_func,
                        resampler='jackknife')
    return fitter.fit(hadron.data, hadron.central_errs,
                      initial_value=initial_value, fit_range=fit_range,
                      covariant=covariant, **kwargs)



from itertools import chain
from pyon.lib.io.folder import get_list_of_files
from pyon.lib.io.formats import parse_iwasaki_32c_charged_meson_file
from pyon.lib.register import Register

registered_parsers = {}


class Parser(object):
    def get_from_file(self, file_name):
        raise NotImplementedError("Implement this in a derived class.")

    def get_from_files(self, list_of_files):
        raw_data = []
        for ff in list_of_files:
                with open(ff, 'r') as f:
                    raw_data.append(self.get_from_file(f))
        to_return = chain.from_iterable(raw_data)
        return to_return

    def get_from_folder(self, folder):
        return self.get_from_files(get_list_of_files(folder))


@Register(registered_parsers, 'iwasaki_32c')
class Iwasaki32c(Parser):
    def get_from_file(self, file_name):
        return parse_iwasaki_32c_charged_meson_file(file_name)





__author__ = 'srd1g10'


class Register(object):
    """
    Decorator to register the class ``cls`` to the dict ``register_to_cls`` with key ``name``. Maintains a list of
    functions that can be used. ::

        >>> from pyon.lib.fitting import registered_fitters
        >>> @Register(registered_fitters, 'scipy')
        ... class ScipyFitter:
        ...     pass
        >>> registered_fitters['scipy']
        <class 'pyon.lib.fitting.ScipyFitter'>

    """
    def __init__(self, register_to, name):
        self.register_to = register_to
        self.name = name

    def __call__(self, cls):
        a = self.register_to  # alias
        if self.name not in a:  # Avoid duplicates
            a.update({self.name: cls})
        return cls




class QuerySet(object):
    def __init__(self, data, manager_class=None):
        self.data = data
        if manager_class is None:
            raise ValueError("Must provide a manager class for a QuerySet")
        self.manager_class = manager_class
        self.manager = manager_class(self)
        self._list_data = None

    def make_query_set(self, data):
        return QuerySet(data, self.manager_class)

    def filter(self, **kwargs):
        return self.make_query_set(self.manager.filter(**kwargs))

    def exclude(self, **kwargs):
        return self.make_query_set(self.manager.exclude(**kwargs))

    def all(self):
        return self.data

    def __getitem__(self, item):
        """
        Allow indexing of data e.g. qs[0]

        Converts into a list to allow indexing and stores this so that it isn't
        calculated twice.
        """
        if self._list_data is None:
            self._list_data = list(self.data)
        return self._list_data[item]

    def sort(self, sort_by):
        self.data = self.manager.sort(sort_by)

    def unique(self, field_name):
        """
        Return a list of the unique values of ``field_name`` in the data.
        """
        return self.manager.unique(field_name)



import json
from numpy import random
from pyon.lib.fitfunction import pp_flat
#from pyon.lib.fitting import fit_hadron
from pyon.lib.fitting import registered_fitters
#from pyon.lib.hadron import Hadron
from pyon.lib.meson import PseudoscalarMeson

__author__ = 'srd1g10'
import unittest


class TestSerialize(unittest.TestCase):
    def setUp(self):
        self.time_extent = 64
        self.fit_func = lambda t, m, c: pp_flat(t, m, c, self.time_extent)

    # def test_serialize_stdout(self):
    #     data = [[pp_flat(t, 1.0, 1.0, self.time_extent) + random.normal(scale=1e-4)
    #              for t in range(self.time_extent)] for _ in range(10)]
    #     had = PseudoscalarMeson(data, masses=(0.01, 0.01))
    #     s = had.dumps(method='string')
    #     print(s)
    #     self.assertTrue(True)

    def test_serialize_json(self):
        data = [[pp_flat(t, 1.0, 1.0, self.time_extent) + random.normal(scale=1e-4)
                 for t in range(self.time_extent)] for _ in range(50)]
        had = PseudoscalarMeson(data, masses=(0.01, 0.01))
        j = had.json()
        had2 = PseudoscalarMeson.from_json(j)
        self.assertEqual(had2.masses, [0.01, 0.01])



import logging
import os
import re

# Matches scientific notation e.g. 1.2345e+06
RE_SCIENTIFIC = "[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?"

IWASAKI_REGEX = {
    'filename': ("\w+\.src\d+\."
                 "ch1(?P<charge1>{res})\.ch2(?P<charge2>{res})\.".format(res=RE_SCIENTIFIC) +
                 "m1(?P<mass1>{res})\.m2(?P<mass2>{res})\.".format(res=RE_SCIENTIFIC) +
                 "dat\.(?P<config_number>\d+)"),
    'data': ("^STARTPROP\n"
             "^MASSES:\s\s(?P<m1>{res})\s{{3}}(?P<m2>{res})".format(res=RE_SCIENTIFIC) + "\n"
             "^SOURCE:\s(?P<source>\w+)\n"
             "^SINKS:\s(?P<sink>\w+)\n"
             "(?P<data>" + "(^\d+\s\s{res}\s\s{res}".format(res=RE_SCIENTIFIC)
             + "\n)+)"
             "^ENDPROP"),
}
IWASAKI_COMPILED_REGEX = re.compile(IWASAKI_REGEX['data'], re.MULTILINE)


def parse_iwasaki_32c_charged_meson_file(f):
    """
    Parse all the data from an Iwasaki Charged Meson file e.g. \
    meson_BOX_RELOADED.src0.ch1-0.3333333333.ch2-0.3333333333.m10.03.m20.03.dat.510
    """
    data = []
    raw_data = f.read()
    fname = os.path.basename(f.name)
    m = re.match(IWASAKI_REGEX['filename'], fname)
    if m:
        charge_1 = int(round(3 * float(m.group('charge1')), 1))
        charge_2 = int(round(3 * float(m.group('charge2')), 1))
        mass_1 = float(m.group('mass1'))
        mass_2 = float(m.group('mass2'))
        config_number = int(m.group('config_number'))
    else:
        raise re.error("Cannot match filename")
    r = IWASAKI_COMPILED_REGEX
    matched = [m.groupdict() for m in r.finditer(raw_data)]
    for match in matched:
        re_data = []
        im_data = []
        time_slices = []
        for line in match['data'].split('\n'):
            try:
                n, re_c, im_c = line.split()
                re_c = float(re_c)
                im_c = float(im_c)
                n = int(n)
                re_data.append(re_c)
                im_data.append(im_c)
                time_slices.append(n)
            except ValueError:
                pass
        dic = {'source': match['source'],
               'sink': match['sink'],
               'data': re_data,
               'im_data': im_data,
               'time_slices': time_slices,
               'masses': (mass_1, mass_2),
               'charges': (charge_1, charge_2),
               'config_number': config_number}
        data.append(dic)
    return data


def filter_correlators(corrs, **kwargs):
    for corr in corrs:
        matched = True
        for k, v in kwargs.items():
            if corr[k] != v:
                matched = False
                break
        if matched:
            return corr
    raise ValueError("Cannot match {}".format(kwargs))


def filter_one_correlator(corr, **kwargs):
    matched = True
    for k, v in kwargs.items():
        if corr[k] != v:
            matched = False
    return matched



import numpy as np
from pyon.lib.fitfunction import effective_mass_pp
from pyon.lib.resampling import Jackknife
import json
import logging


class Hadron(object):
    """
    :class:`Hadron` is a representation of a QCD hadron. A particular class
    that inherits from this will specify the ``fit_func`` and how to plot the
    data that represents a certain correlation function.
    """
    def __init__(self, data=None, config_numbers=None, sort=True, source=None,
                 sink=None, im_data=None,
                 time_slices=None):
        self.data = data
        if sort and config_numbers:
            self.config_numbers = config_numbers
            self.sort()
        else:
            self.config_numbers = [-1 for _ in self.data]
        self.time_extent = len(data[0])  # Assume it's the same for samples
        self.fit_func = None  # define this in the derived class
        self.source = source
        self.sink = sink
        self.im_data = im_data
        self.time_slices = time_slices

    @classmethod
    def from_parsed_data(cls, parsed_data):
        """
        Create a :class:`Hadron` from ``parsed_data``. This handles all the \
        duplicate information that the parsed data
        might have. The ``parsed_data`` will be the return value from e.g. \
        :func:`parse_iwasaki_32c_charged_meson_file \
        <.io.formats.parse_iwasaki_32c_charged_meson_file>`
        """
        data = [fd['data'] for fd in parsed_data]
        config_numbers = [fd['config_number'] for fd in parsed_data]
        kwargs = parsed_data[0]  # assume all parameters are the same as the first one
        kwargs.pop('data')
        kwargs.pop('config_number')
        kwargs['config_numbers'] = config_numbers
        return cls(data, **kwargs)

    @classmethod
    def from_json(cls, json_data):
        """
        Create :class:`Hadron` from JSON data. Defaults to passing arguments to
        the default constructor.
        """
        return cls(**json.loads(json_data))

    @classmethod
    def from_view(cls, view):
        """

        """
        return cls.from_parsed_data(view.data)

    def __str__(self):
        raise NotImplementedError

    def _name(self):
        raise NotImplementedError

    def dump(self, **kwargs):
        """
        Can override this to call some other serializer to dump e.g. Pickle/raw
         string
        """
        return self.json(**kwargs)

    def json(self, **kwargs):
        """
        Returns the JSON encoded representation of the hadron.

        :param \*\*kwargs: Optional arguments that :func:`json.dumps` takes.
        """
        to_dump = self._to_dump()
        return json.dumps(to_dump, **kwargs)

    def _to_dump(self):
        """
        Specify the fields that will be dumped by :func:`json`.
        See :func:`ScipyFitter._to_dump` for an example of how to inherit and
        adapt this function.
        """
        return {'data': self.data, 'config_numbers': self.config_numbers, 'source': self.source, 'sink': self.sink,
                'im_data': self.im_data, 'time_slices': self.time_slices}

    def sort(self):
        """
        Sorts the data using ``config_numbers`` as a key.
        """
        self.data = [x for (y, x) in sorted(zip(self.config_numbers,
                                                self.data))]
        self.config_numbers.sort()

    def scale(self):
        """
        Scales the data using ``self.central_data[0]`` as a scale factor.
        Scaling can improve the convergence of fitting since we aren't dealing
        with huge numbers.
        """
        scale_factor = self.central_data[0]
        self.data = [[P / scale_factor for P in c] for c in self.data]

    def fold(self):
        """
        Fold the data to improve statistics. Defaults to PP folding i.e. fold
        down the middle time slice.
        """
        t_ext = self.time_extent
        self.data = [self._fold_one(c, t_ext) for c in self.data]

    @staticmethod
    def _fold_one(corr, t_ext):
        return [0.5*(corr[t] + corr[(t_ext-t) % t_ext]) for t in range(t_ext)]

    @property
    def central_data(self):
        return np.average(self.data, axis=0)

    @property
    def central_errs(self):
        return np.std(self.data, axis=0) / len(self.config_numbers)

    @property
    def effective_mass(self):
        return self._effective_mass_fn(self.central_data)

    @property
    def effective_mass_errs(self):
        """
        Return the central errors on the effective mass. By default this will
        calculate them via jackknife.
        """
        resampler = Jackknife(n=1)
        samples = [self._effective_mass_fn(j) for j in resampler.generate_samples(self.data)]
        errs = [resampler.calculate_errors(cent, col) for col, cent in zip(columns(samples), self.effective_mass)]
        return errs

    @staticmethod
    def _effective_mass_fn(data):
        return effective_mass_pp(data)


def columns(mat):
    for n in range(len(mat[0])):
        yield np.array(mat)[:, n]



import os
from random import shuffle
from numpy import random
from pyon.lib.fitfunction import pp_flat
from pyon.lib.fitting import registered_fitters, fit_hadron
from pyon.lib.io.folder import get_list_of_files
from pyon.lib.io.formats import parse_iwasaki_32c_charged_meson_file, \
    filter_correlators
from pyon.lib.meson import PseudoscalarMeson, PseudoscalarChargedMeson

__author__ = 'srd1g10'
import unittest


class TestHadron(unittest.TestCase):
    def setUp(self):
        self.time_extent = 64
        self.fit_func = lambda t, m, c: pp_flat(t, m, c, self.time_extent)
        self.iwasaki_folder = os.path.join('testfiles', 'parse', 'f1')

    def test_make_hadron(self):
        data = [[pp_flat(t, 1.0, 1.0, self.time_extent) + random.normal(scale=1e-4)
                 for t in range(self.time_extent)] for _ in range(10)]
        had = PseudoscalarMeson(data, masses=(0.01, 0.01))
        self.failUnlessEqual(had.data[0][0], data[0][0])
        self.failUnlessEqual(had.masses, (0.01, 0.01))

    #@unittest.skip('slow')
    def test_fit_hadron(self):
        data = [[pp_flat(t, 1.0, 1.0, self.time_extent) + random.normal(scale=1e-4)
                 for t in range(self.time_extent)] for _ in range(50)]
        had = PseudoscalarMeson(data, masses=(0.01, 0.01))
        fp = fit_hadron(had, fit_range=range(1, 32+1),
                        initial_value=[0.8, 1.5], covariant=False)
        self.assertAlmostEqual(fp.average_params['m'], 1.0, 2)

    def make_filtered_data(self):
        raw_data = []
        for ff in get_list_of_files(self.iwasaki_folder):
            with open(ff, 'r') as f:
                raw_data.append(parse_iwasaki_32c_charged_meson_file(f))
        filtered_data = [filter_correlators(rd, source='GAM_5', sink='GAM_5', masses=(0.03, 0.03))
                         for rd in raw_data]
        return filtered_data

    def fit_test_hadron(self):
        filtered_data = self.make_filtered_data()
        had = PseudoscalarChargedMeson.from_parsed_data(filtered_data)
        had.sort()
        had.scale()
        fp = fit_hadron(had, fit_range=range(1, 32+1), initial_value=[0.3212, 1.65], covariant=False)
        return fp

    def test_sort(self):
        filtered_data = self.make_filtered_data()
        shuffle(filtered_data)  # shuffle to test sort
        shuffled_config_numbers = [fd['config_number'] for fd in filtered_data]
        had = PseudoscalarChargedMeson.from_parsed_data(filtered_data)
        had.sort()
        had.scale()
        self.assertEqual(had.config_numbers[0], 510)
        self.assertNotEqual(had.config_numbers, shuffled_config_numbers)

    def test_scale(self):
        filtered_data = self.make_filtered_data()
        had = PseudoscalarChargedMeson.from_parsed_data(filtered_data)
        had.sort()
        had.scale()
        self.assertAlmostEqual(had.central_data[0], 1.0, 5)

    def test_fold(self):
        filtered_data = self.make_filtered_data()
        had = PseudoscalarChargedMeson.from_parsed_data(filtered_data)
        had.sort()
        had.scale()
        had.fold()
        self.assertEqual(had.data[0][1], had.data[0][-1])


"""
The point of this module is to be able to easily plot lattice QCD things - most often this will be correlators.
We want to have the same syntax for common use plotting regardless of what program is used to plot it. So, the classes
here implement a common API that forwards stuff to e.g. Pyplot and Gnuplot
"""
import matplotlib.pyplot as plt
#from pyon.lib.fitfunction import pp_effective_mass
from pyon.lib.register import Register
import numpy as np
__author__ = 'srd1g10'

registered_plotters = {}


class Plotter(object):
    """
    A :class:`Plotter` plots hadrons like mesons, baryons. e.g. ::

        >>> from pyon.lib.meson import PseudoscalarMeson
        >>> from pyon.lib.fitfunction import pp_flat
        >>> data = [
        ...    [pp_flat(t, 1.0, 1.0, 64) for t in range(64)]
        ...    for _ in range(200) ]
        >>> had = PseudoscalarMeson(data, masses=(0.05, 0.05))
        >>> plot_hadron(had)

    """
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def plot(self, *args, **kwargs):
        pass

    def show(self):
        pass


@Register(registered_plotters, 'pyplot')
class PyplotPlotter(Plotter):
    """
    Uses matplotlib.pyplot to do the plotting.
    """

    def __init__(self, *args, **kwargs):
        super(PyplotPlotter, self).__init__(*args, **kwargs)
        self.fig, self.ax = plt.subplots()

    def __exit__(self, exc_type, exc_val, exc_tb):
        plt.close('all')
        return False

    def plot(self, x, y, yerr=None, title=None, log_scale=True, *args, **kwargs):
        """
        Set up the fields of the object to the specifics of the plot.
        """
        if title is not None:
            self.ax.set_title(title)
        self.ax.grid(True)
        self.ax.errorbar(x, y, yerr=yerr, *args, **kwargs)

    def set_scale(self, x_scale=None, y_scale=None):
        if x_scale:
            self.ax.set_xscale(x_scale)
        if y_scale:
            self.ax.set_yscale(y_scale)

    def show(self):
        self.fig.show()


def plot_hadron(hadron, title=None, effective_mass=False, **kwargs):
    """
    Simple common use case. Defaults to PyplotPlotter.

    :param hadron: The :class:`Hadron <hadron.Hadron>` object to plot.
    :param \*\*kwargs: The keyword arguments to pass to :func:`matplotlib.pyplot.errorbar`.
    """
    plotter = PyplotPlotter()
    if effective_mass:
        data = hadron.effective_mass
        errs = hadron.effective_mass_errs
        y_scale = None
    else:
        data = hadron.central_data
        errs = hadron.central_errs
        y_scale = 'log'
    t_range = range(hadron.time_extent//2)
    data = [data[t] for t in t_range]
    errs = [errs[t] for t in t_range]
    if title is None:
        title = str(hadron)
    plotter.plot(t_range, data, yerr=errs, title=title, **kwargs)
    plotter.set_scale(y_scale=y_scale)
    return plotter





from numpy import random
import os
from pyon.lib.fitfunction import pp_flat
from pyon.lib.fitting import ScipyFitter, fit_hadron
from pyon.lib.io.folder import get_list_of_files
from pyon.lib.io.formats import parse_iwasaki_32c_charged_meson_file, \
    filter_correlators
from pyon.lib.meson import PseudoscalarChargedMeson
from pyon.lib.resampling import Jackknife

__author__ = 'srd1g10'
import unittest


class TestFitting(unittest.TestCase):
    def setUp(self):
        """
        Using (0.03, 0.03), (-1, -1) from 0.004 32c
        """
        self.time_extent = 64
        self.corr = [pp_flat(t, 1.0, 1.0, self.time_extent) for t in range(self.time_extent)]
        self.corrs = [[pp_flat(t, 1.0, 1.0, self.time_extent) + random.normal(scale=1e-6)
                       for t in range(self.time_extent)] for _ in range(200)]
        self.iwasaki_folder = os.path.join('testfiles', 'parse', 'f1')
        self.fit_func = lambda t, m, c: pp_flat(t, m, c, self.time_extent)

    def test_make_correlator(self):
        corr = [pp_flat(t, 1.0, 1.0, self.time_extent) for t in range(self.time_extent)]
        self.failUnlessEqual(corr[0], 1.)
        self.failIfEqual(corr[0], 2.)

    def test_fit_one_correlator(self):
        fitter = ScipyFitter(fit_func=self.fit_func)
        fit_params = fitter.fit_one(self.corr,
                                    fit_range=range(1, 32+1),
                                    initial_value=[1.3, 1.0])
        self.failUnless(fit_params is not None)  # fail if the fit fails or fit_params don't exist
        mass = fit_params['m']
        self.failUnlessAlmostEqual(mass, 1.0, delta=1e-3)
        self.failIfAlmostEqual(mass, 2.0, delta=1e-3)

    def test_fail_fit(self):
        fitter = ScipyFitter()
        self.assertRaises(TypeError, fitter.fit_one, self.corr,
                          fit_range=range(1, 32+1),
                          initial_value=[0.8, 1.5])

    def make_hadron(self):
        raw_data = []
        for ff in get_list_of_files(self.iwasaki_folder):
            with open(ff, 'r') as f:
                raw_data.append(parse_iwasaki_32c_charged_meson_file(f))
        filtered_data = [filter_correlators(rd, source='GAM_5', sink='GAM_5', masses=(0.03, 0.03))
                         for rd in raw_data]
        had = PseudoscalarChargedMeson.from_parsed_data(filtered_data)
        had.sort()
        had.fold()
        had.scale()
        return had

    def test_fit_covar(self):
        had = self.make_hadron()
        bnds = ((0., 1.), (0, None))
        fp = fit_hadron(had, fit_range=range(9, 32+1), initial_value=[0.3212, 1.654], covariant=True, bounds=bnds)
        fit_params = fp.average_params
        self.assertTrue(fit_params is not None)  # fail if the fit fails or fit_params don't exist
        mass = fit_params['m']
        c = fit_params['c']
        self.failUnlessAlmostEqual(mass, 0.32120951002753384, delta=1e-6)
        self.failUnlessAlmostEqual(c, 1.6542423508883226, delta=1e-6)

    def test_fit_covar_err(self):
        had = self.make_hadron()
        bnds = ((0., 1.), (0., None))
        fp = fit_hadron(had, fit_range=range(9, 32+1), initial_value=[0.3212, 1.654], covariant=True, bounds=bnds)
        errs = fp.errs
        self.assertTrue(errs is not None)  # fail if the fit fails or fit_params don't exist
        mass = errs['m']
        c = errs['c']
        self.failUnlessAlmostEqual(mass, 0.0003960453177452647, delta=1e-2)
        self.failUnlessAlmostEqual(c, 0.014091062055500116, delta=1e-2)

    def test_fit(self):
        fitter = ScipyFitter(fit_func=self.fit_func)
        f = fitter.fit(self.corrs,
                       fit_range=range(1, 32+1),
                       initial_value=[1.4, 1.5], covariant=False)
        fit_params = f.average_params
        self.assertTrue(fit_params is not None)  # fail if the fit fails or fit_params don't exist
        mass = fit_params['m']
        self.failUnlessAlmostEqual(mass, 1.0, delta=1e-2)
        self.failIfAlmostEqual(mass, 2.0, delta=1e-3)


from pyon import registered_simulations, registered_views
from pyon import registered_models
from pyon import registered_sources

__author__ = 'srd1g10'


def create_register(registry):
    class Register(object):
        def __init__(self, name):
            self.register_to = registry
            self.name = name

        def __call__(self, cls):
            a = self.register_to  # alias
            if self.name not in a:  # Avoid duplicates
                a.update({self.name: cls})
            return cls
    return Register


view = create_register(registered_views)

source = create_register(registered_sources)

model = create_register(registered_models)

simulation = create_register(registered_simulations)


registered_views = {}


class View(object):
    queryset = None
    data_source = None
    filter = None

    def get_queryset(self):
        if self.queryset is None:
            if self.filter is not None:
                return self.data_source.filter(**self.filter)
            else:
                return self.data_source.all()
        else:
            return self.queryset

    @property
    def data(self):
        return self.get_queryset().all()




import json
from pyon.lib.register import Register

registered_serializers = {}


class Serializer(object):
    def dump(self):
        pass

    def dumps(self):
        pass

    def load(self):
        pass

    def loads(self):
        pass


@Register(registered_serializers, 'json')
class JSONSerializer(Serializer):
    def dump(self, *args, **kwargs):
        json.dump(*args, **kwargs)

    def dumps(self, *args, **kwargs):
        json.dumps(*args, **kwargs)

    def load(self, *args, **kwargs):
        json.load(*args, **kwargs)

    def loads(self, *args, **kwargs):
        json.loads(*args, **kwargs)


@Register(registered_serializers, 'string')
class StringSerializer(Serializer):
    """
    Just return a string.
    """
    def dumps(self, *args):
        return args[0]


import math
import numpy as np
from pyon.lib.register import Register

__author__ = 'srd1g10'


registered_resamplers = {}


@Register(registered_resamplers, 'jackknife')
class Jackknife:
    """
    Jackknife resampler.

    :param n: The ``n`` parameter in the jackknife method.
    :type n: int
    """
    def __init__(self, n=1):
        self.n = n

    def generate_samples(self, data):
        length = len(data)
        lists = []
        selectevery = int(length/self.n)
        for j in range(selectevery):
            this_list = []
            for i in range(length):
                if i % selectevery != j:
                    this_list.append(data[i])
            lists.append(this_list)
        return np.average(lists, axis=1)

    def calculate_fit_errors(self, central, samples):
        fit_errs = {}
        for k, central in central.items():
            fit_errs[k] = self._jackknife_error(central, samples[k])
        return fit_errs

    def calculate_errors(self, central, samples):
        return self._jackknife_error(central, samples)

    @staticmethod
    def _jackknife_error(c, samples):
        if not isinstance(samples, np.ndarray):
            samples = np.array(samples)
        s = 0.0
        for l in samples:
            d = c-l
            if math.isnan(d):  # Ignore infinities
                d = 0.0
            s += d*d
        s *= (1.0 - 1.0 / len(samples))
        return np.sqrt(s)


@Register(registered_resamplers, 'none')
class NoResampler:
    """
    Doesn't do any resampling (for testing mainly)
    """
    @staticmethod
    def generate_samples(data):
        return data

    @staticmethod
    def calculate_errors(average_params, resampled_params):
        return np.std(resampled_params)

# def jackknife_reduce(data, n=1):
#     """Returns averaged jackknife lists"""
#     lists = [np.average(l, axis=0) for l in generate_jackknife_samples(data, n)]
#     return lists



from numpy import random
from pyon.lib.fitfunction import pp_flat
from pyon.lib.meson import PseudoscalarMeson
from pyon.lib.plotting import plot_hadron
import unittest


class TestPlot(unittest.TestCase):
    def setUp(self):
        self.time_extent = 64

    @unittest.skip('annoying')
    def test_plot_hadron(self):
        data = [[pp_flat(t, 1.0, 1.0, self.time_extent) + random.normal(scale=1e-9)
                 for t in range(self.time_extent)] for _ in range(10)]
        had = PseudoscalarMeson(data, masses=(0.01, 0.01))
        plt = plot_hadron(had)
        plt.show()

    @unittest.skip('annoying')
    def test_eff_mass_hadron(self):
        data = [[pp_flat(t, 1.0, 1.0, self.time_extent) + random.normal(scale=1e-9)
                 for t in range(self.time_extent)] for _ in range(10)]
        had = PseudoscalarMeson(data, masses=(0.01, 0.01))
        plt = plot_hadron(had, effective_mass=True)
        plt.show()



import functools
from pyon.lib.fitfunction import pp_flat
from pyon.lib.hadron import Hadron


class Meson(Hadron):
    def __init__(self, data=None, config_numbers=None, masses=None, **kwargs):
        """
        Creates a meson.

        :param data: The measurements of the correlation functions.
        :type data: list of floats
        :param masses: The bare masses of the quarks of the meson.
        :type masses: tuple of floats e.g. (0.01, 0.005)
        """
        super(Meson, self).__init__(data, config_numbers, **kwargs)
        self.masses = masses

    def _to_dump(self):
        to_dump = super(Meson, self)._to_dump()
        to_dump.update({'masses': self.masses})
        return to_dump

    def __str__(self):
        return "{}, Masses: {}".format(self._name, self.masses)

    @property
    def _name(self):
        return "Meson"


class PseudoscalarMeson(Meson):
    """
    A :class:`PseudoscalarMeson` is a :class:`Hadron <hadron.Hadron>`
    consisting of two valence quarks.
    """

    def __init__(self, data=None, config_numbers=None, masses=None, **kwargs):
        """
        Creates an uncharged pseudoscalar meson.

        :param data: The measurements of the correlation functions.
        :type data: list of floats
        :param masses: The bare masses of the quarks of the meson.
        :type masses: tuple of floats e.g. (0.01, 0.005)
        """
        super(PseudoscalarMeson, self).__init__(data, config_numbers, masses,
                                                **kwargs)
        #self.fit_func = functools.partial(pp_flat, T=self.time_extent)
        self.fit_func = lambda t, m, c: pp_flat(t, m, c,
                                                      self.time_extent)
        #self.hess = lambda t, m, c: pp_flat_hess(t, m, c, self.time_extent)

    @property
    def _name(self):
        return "Pseudoscalar Meson"


class PseudoscalarChargedMeson(Meson):
    def __init__(self, data=None, config_numbers=None, masses=None,
                 charges=None, **kwargs):
        super(PseudoscalarChargedMeson, self).__init__(data, config_numbers,
                                                       masses, **kwargs)
        self.charges = charges
        self.fit_func = lambda t, m, c: pp_flat(t, m, c,
                                                      self.time_extent)
        #self.hess = lambda t, m, c: pp_flat_hess(t, m, c, self.time_extent)

    def _to_dump(self):
        to_dump = super(PseudoscalarChargedMeson, self)._to_dump()
        to_dump.update({'charges': self.charges})
        return to_dump

    def __str__(self):
        return super(PseudoscalarChargedMeson, self).__str__() + \
            ", Charges: {}".format(self.charges)

    @property
    def _name(self):
        return "Pseudoscalar Charged Meson"



import os
__author__ = 'srd1g10'


def sorted_ls(path):
    """
    Sort by name
    """
    return list(sorted(os.listdir(path)))


def get_list_of_files(folder):
    """
    Returns list of files in folder with the folder prepended.
    """
    out = sorted_ls(folder)
    if out:
        list_of_files = [os.path.join(folder, f) for f in out]
        return list_of_files
    else:
        raise IOError("Error opening {0}, skipping.".format(folder))


__author__ = 'srd1g10'
import numpy as np


def get_cov(X, Y):
    """
    Returns covariance of the two vectors ``X`` and ``Y``.
    Formula is: :math:`Cov(X, Y) = 1/(n - 1) \sum_{i = 1}^{n} (x_i - \hat x) (y_i - \hat y)`

    :param X: Vector of floats
    :param Y: Vector of floats
    :rtype: float
    """
    N = len(X)
    mean_x = np.average(X)
    mean_y = np.average(Y)
    total = np.sum([(x - mean_x)*(y - mean_y) for x,y in zip(X,Y)])
    cov = 1.0 / (N - 1.0) * total
    return cov


def get_inverse_cov_matrix(M, correlated=False):
    """
    Returns the inverse covariance matrix of ``M``.

    :param M: The matrix of floats
    :param correlated: Whether the calculation uses the correlated form or not.
    :type correlated: bool
    :rtype: :class:`numpy.matrix`
    """
    Ndata = len(M)
    Nconf = len(M[0])
    cov_matrix = np.zeros((Ndata, Ndata))
    for i in range(Ndata):
        row_i = M[i,:]
        for j in range(Ndata):
            row_j = M[j,:]
            if i != j and not correlated:
                continue
            temp = get_cov(row_i, row_j) / Nconf
            cov_matrix[i][j] = temp
    cov_matrix = np.matrix(cov_matrix)
    invcov = cov_matrix.I
    return invcov


from random import shuffle
from pyon.lib.fitting import fit_hadron

__author__ = 'srd1g10'
import os
from pyon.lib.io.folder import get_list_of_files
from pyon.lib.io.formats import parse_iwasaki_32c_charged_meson_file, \
    filter_correlators
from pyon.lib.meson import PseudoscalarChargedMeson
import unittest


class TestParsing(unittest.TestCase):
    def setUp(self):
        self.iwasaki_file = os.path.join('testfiles', 'parse', 'f1',
                                         'meson_BOX_RELOADED.src0.ch1-0.3333333333.'
                                         'ch2-0.3333333333.m10.03.m20.03.dat.510')
        self.iwasaki_folder = os.path.join('testfiles', 'parse', 'f1')

    def test_parse_iwasaki(self):
        with open(self.iwasaki_file, 'r') as f:
            raw_data = parse_iwasaki_32c_charged_meson_file(f)
        filtered_data = filter_correlators(raw_data, source='GAM_5', sink='GAM_5', masses=(0.03, 0.03))
        self.failUnlessEqual(filtered_data['data'][0], 4.984743e+06)
        self.failUnlessEqual(filtered_data['masses'], (0.03, 0.03))
        self.failUnlessEqual(filtered_data['source'], 'GAM_5')
        self.failUnlessEqual(filtered_data['config_number'], 510)

    def test_fail_parse_iwasaki(self):
        with open(self.iwasaki_file, 'r') as f:
            raw_data = parse_iwasaki_32c_charged_meson_file(f)
        self.assertRaises(ValueError, filter_correlators, raw_data, source='GAM_5', sink='dfjkdfjk',
                          masses=(0.03, 0.03))

    def test_parse_multiple_iwasaki(self):
        raw_data = []
        for ff in get_list_of_files(self.iwasaki_folder):
            with open(ff, 'r') as f:
                raw_data.append(parse_iwasaki_32c_charged_meson_file(f))
        filtered_data = [filter_correlators(rd, source='GAM_5', sink='GAM_5', masses=(0.03, 0.03)) for rd in raw_data]
        filtered_data.sort(key=lambda k: k['config_number'])
        self.assertEqual(filtered_data[0]['data'][0], 4.984743e+06)
        self.assertEqual(filtered_data[0]['masses'], (0.03, 0.03))
        self.assertEqual(filtered_data[0]['source'], 'GAM_5')
        self.assertEqual(filtered_data[0]['config_number'], 510)
        self.assertNotEqual(filtered_data[1]['config_number'], 510)

    def make_filtered_data(self):
        raw_data = []
        for ff in get_list_of_files(self.iwasaki_folder):
            with open(ff, 'r') as f:
                raw_data.append(parse_iwasaki_32c_charged_meson_file(f))
        filtered_data = [filter_correlators(rd, source='GAM_5', sink='GAM_5', masses=(0.03, 0.03))
                         for rd in raw_data]
        return filtered_data

    def fit_test_hadron(self):
        filtered_data = self.make_filtered_data()
        had = PseudoscalarChargedMeson.from_parsed_data(filtered_data)
        had.sort()
        had.scale()
        fp = fit_hadron(had, fit_range=range(1, 32+1), initial_value=[0.3212, 1.65], covariant=False)
        return fp


    @unittest.skip('slow')
    def test_fit_parsed_data(self):
        fit_params = self.fit_test_hadron().average_params
        self.assertTrue(fit_params is not None)  # fail if the fit fails or fit_params don't exist
        mass = fit_params['m']
        c = fit_params['c']
        self.failUnlessAlmostEqual(mass, 0.3212, delta=1e-2)
        self.failUnlessAlmostEqual(c, 1.53, delta=1e-2)

    @unittest.skip('slow')
    def test_fit_parsed_error(self):
        fit_errors = self.fit_test_hadron().errs
        self.assertTrue(fit_errors is not None)  # fail if the fit fails or fit_params don't exist
        m_err = fit_errors['m']
        c_err = fit_errors['c']
        self.failUnlessAlmostEqual(m_err, 0.00044, delta=1e-4)
        self.failUnlessAlmostEqual(c_err, 0.0053, delta=1e-4)



import logging
from pyon.lib.manager import DBManager, FileManager
from pyon.runner.query import QuerySet
from pyon import registered_parsers
registered_sources = {}


# class Source2:
#     data_format = None
#     folder = None
#     files = None
#     data_format_kwargs = {}
#     _data = None
#
#     def __init__(self, manager=None):
#         self.manager = manager  # or DefaultManager()
#
#     def get_data(self):
#         parser = self.get_parser()
#         if self.files:
#             return parser.get_from_files(self.files)
#         elif self.folder:
#             return parser.get_from_folder(self.folder)
#
#     def get_parser(self):
#         return registered_parsers[self.data_format](**self.data_format_kwargs)
#
#     def all(self):
#         return self.get_queryset()
#
#     def get_queryset(self):
#         if self._data is None:
#             logging.debug("Getting data from {}".format(self.folder))
#             self._data = self.get_data()
#         return QuerySet(self._data)


class Source:
    def __init__(self, manager=None):
        self.manager = manager
        self._data = None

    @property
    def data(self):
        if self._data is None:
            self._data = self._get_data()
        return self._data

    def filter(self, **kwargs):
        qs = QuerySet(self.data, self.manager)
        return qs.filter(**kwargs)

    def exclude(self, **kwargs):
        qs = QuerySet(self.data, self.manager)
        return qs.exclude(**kwargs)

    def all(self):
        return QuerySet(self.data, self.manager).all()

    def _get_data(self):
        pass


class FileSource(Source):
    data_format = None
    folder = None
    files = None

    def __init__(self):
        super(FileSource, self).__init__(FileManager)

    def _get_data(self):
        parser = self._get_parser()
        if self.files:
            return parser.get_from_files(self.files)
        elif self.folder:
            return parser.get_from_folder(self.folder)

    def _get_parser(self):
        return registered_parsers[self.data_format]()


class DBSource(Source):
    def __init__(self):
        super(DBSource, self).__init__(DBManager)



# Needed? There's nothing here!
class Model(object):
    """
    A :class:`Model` defines the various formulae of a particular problem that
    we're interested in. The methods use a :class:`View`, which is a particular
    aspect of some data source.
    """
registered_models = {}


__author__ = 'srd1g10'
from functools import partial
import numpy as np


def pp(t, m, z, T=64):
    return 0.5*z*z/m*(np.exp(-m*t) + np.exp(-m*(T-t)))


def aa(t, m, f, Z, T=64):
    return 0.5*f*f*m/(Z*Z)*(np.exp(-m*t) + np.exp(-m*(T-t)))


def ap(t, m, z, f, Z, T=64):
    return 0.5*z*f/Z*(np.exp(-m*t) - np.exp(-m*(T-t)))


def pp_flat(t=1, m=1.0, c=1.0, T=64):
    return c * (np.exp(-m*t) + np.exp(-m*(T-t)))


# def pp_flat_hess(t=1, m=1.0, c=1.0, T=64):
#     h_mm = c * (t * t * np.exp(-m*t) + (T-t)*(T-t) * np.exp(-m*(T-t)))
#     h_mc = -(t * np.exp(-m*t) + (T-t)*np.exp(-m*(T-t)))
#     h_cm = h_mc
#     h_cc = 0.
#     return [[h_mm, h_mc], [h_cm, h_cc]]


def pp_flat_hess(t=1, m=1.0, c=1.0, T=64):
    h_m = -c * (t * np.exp(-m*t) + (T-t)*np.exp(-m*(T-t)))
    h_c = (np.exp(-m*t) + np.exp(-m*(T-t)))
    return [h_m, h_c]


def effective_mass_pp(corr):
    np.seterr('ignore')
    t_ext = len(corr)
    eff = [np.arccosh((corr[(t-1) % t_ext] + corr[(t + 1) % t_ext])/(2. * corr[t % t_ext])) for t in range(t_ext)]
    np.seterr('warn')
    return eff



from pyon.lib.io.formats import filter_one_correlator


class Manager(object):
    """
    A manager defines how to access and filter a :class:`Source`.
    """

    def __init__(self, query_set):
        self.qs = query_set

    def filter(self, **kwargs):
        pass

    def exclude(self, **kwargs):
        pass

    def sort(self, sort_by):
        pass

    def unique(self, field_name):
        pass


class DBManager(Manager):
    def filter(self, **kwargs):
        pass

    def exclude(self, **kwargs):
        pass

    def filter_to_query(self, filter):
        pass

    def _do_query(self, query):
        pass


class FileManager(Manager):
    def filter(self, **kwargs):
        my_filter = lambda x: filter_one_correlator(x, **kwargs)
        filtered = self._do_filter(my_filter, self.qs.all())
        if len(filtered) == 0:
            raise ValueError("Cannot match {}".format(kwargs))
        return filtered

    def exclude(self, **kwargs):
        my_filter = lambda x: not filter_one_correlator(x, **kwargs)
        filtered = self._do_filter(my_filter, self.qs.all())
        if len(filtered) == 0:
            raise ValueError("Cannot match {}".format(kwargs))
        return filtered

    @staticmethod
    def _do_filter(filt, data):
        filtered = list(filter(filt, data))
        return filtered

    def sort(self, sort_by):
        return sorted(list(self.qs.all()), key=lambda item: item[sort_by])

    def unique(self, field_name):
        values = []
        for f in self.qs.all():
            try:
                v = f[field_name]
                if v not in values:
                    values.append(v)
            except KeyError:
                pass
        return values



__author__ = 'srd1g10'

registered_simulations = {}


class Simulation(object):
    """
    A simulation combines a view (aspect of data) and a model to perform some
    calculation with the data and the formulae of the model.
    """
    def __init__(self, model, view):
        self.model = model
        self.view = view
        self.simulation_params = None
        self.simulation_results = None

    def set_simulation_params(self, **kwargs):
        self.simulation_params = kwargs

    def do_simulation(self):
        raise NotImplementedError("Implement this in a derived class.")

    def get_results(self):
        if self.simulation_results is None:
            self.simulation_results = self.do_simulation()
        return self.simulation_results



import imp
import simplejson
import os
import logging
import time


class App(object):
    def __init__(self, directory, name=None, dump_dir=None, template=None,
                 db_path=None):
        self.name = name or time.strftime("%Y%m%d-%H%M%S")
        self.registers = {}
        self.module_names = ['sources', 'views', 'models', 'simulations']
        self._populate_registers(directory)
        self.simulation_results = {}
        self.dump_dir = dump_dir
        self.template = template
        self.db_path = db_path
        self.report_name = 'report.html'
        self.dump_name = 'dump.json'

    def _populate_registers(self, directory):
        """
        Populates the registers with anything that used the appropriate
        decorator.
        """
        for mod in self.module_names:
            imp.load_source(mod, os.path.join(directory, mod + '.py'))
        #  Now the modules are loaded and the relevant registers are populated.
        from pyon import registered_sources, registered_views, \
            registered_models, registered_simulations

        self.registers['sources'] = registered_sources
        self.registers['views'] = registered_views
        self.registers['models'] = registered_models
        self.registers['simulations'] = registered_simulations

    def main(self):
        logging.debug("Running App: {}".format(self.name))
        for sim_name, sim in self.registers['simulations'].items():
            logging.info("Doing {}".format(sim_name))
            my_sim = sim()
            sim_results = my_sim.get_results()
            self.simulation_results[sim_name] = sim_results

            if self.dump_dir:
                self.dump_result(sim_name, sim_results)

            if self.template:
                self.write_report(sim_name, sim_results)

    def dump_result(self, sim_name, sim_results):
        dump_path = os.path.join(self.dump_dir,
                                 self.name
                                 + '_' + sim_name)
        if not os.path.exists(dump_path):
            os.mkdir(dump_path)
        logging.info("Dumping results to {}".format(dump_path))
        to_dump = self.dumps(sim_name, sim_results)
        with open(os.path.join(dump_path, self.dump_name), 'w') as f:
            simplejson.dump(to_dump, f, namedtuple_as_object=True)

    def write_report(self, sim_name, sim_results):
        rendered = self.render_template(self.template, sim_name,
                                        sim_results)
        report_dir = os.path.join(self.dump_dir,
                                  self.name + '_' + sim_name)
        if not os.path.exists(report_dir):
            os.mkdir(report_dir)
        with open(os.path.join(report_dir, self.report_name), 'w') as f:
            f.write(rendered)

    def dumps(self, sim_name, results):
        to_dump = self._sanitize_dic(results)
        to_dump['date'] = time.strftime("%c")
        to_dump['sim_name'] = sim_name
        return to_dump

    @staticmethod
    def render_template(template, sim_name, results):
        template_dic = {
            'title': 'Report: ' + sim_name,
            'sim_results': results,
        }
        rendered = template.render(**template_dic)
        return rendered

    @staticmethod
    def _sanitize_dic(dic):
        new_dic = {}
        for key in dic.keys():
            if type(key) is not str:
                try:
                    new_dic[str(key)] = dic[key]
                except:
                    try:
                        new_dic[repr(key)] = dic[key]
                    except:
                        logging.debug("Sanitize failed for {}".format(key))
                        pass
            else:
                new_dic[key] = dic[key]
        return new_dic
